#
# Copyright (c) 2006 Mike L. Austin.  All rights reserved.
# see LICENCE.txt for details
#

require 'geometry'
require 'events'
require 'uistyle'

module Inertia

module ViewEvents
  def mouse_down( button, origin ) end
  def mouse_up( button, origin ) end
  def mouse_move( origin ) end
  def mouse_drag( origin ) end
  def mouse_enter( origin ) end
  def mouse_exit( origin ) end
  def drag_accept( origin, object ) end
  def drag_drop( object ) end
  def view_size( delta ) end
  def key_down( keysym, value ) end
  def focus_lost() end
end

module ViewRect
  def left() origin.x end;
  def left=( value ) origin.x = value; origin_changed() end
  def top() origin.y end;
  def top=( value ) origin.y = value; origin_changed() end
  def width() extent.x end;
  def width=( value ) extent.x = value; extent_changed( value, height ) end
  def height() extent.y end;
  def height=( value ) extent.y = value; extent_changed( width, value ) end
end

class View < Object

  @source = __FILE__
  @fields = [:origin, :extent, :reshape, :back_color]

  @@graphics = Graphics.new 

  def View.fields()
    @fields + (superclass != Object ? superclass.fields : [])
  end
  
  def View.source()
    @source
  end

  attr :parent, true
  attr :subviews
  attr :focused_view, true
  attr :origin
  attr :extent
  attr :min_size, true
  attr :reshape, true
  attr :uistyle, true
  attr :back_color, true
  attr :border_color, true

  # ------------------------------------------------------------------------ #
  class General; end
  # ------------------------------------------------------------------------ #

  def View.inherited( subclass )
    subclass.class_eval do
      @source = ""
      @fields = []
    end
  end
  
  def initialize()
    super
    @subviews = Array.new
    @origin = Point.new
    @extent = Point.new( 100, 100 )
    @min_size = Point.new( 10, 10 )
    @reshape = [nil, nil]
    @shelves = []
    @mouse_hover = false
    @style = nil
    @screen_origin = Point.new()
    class_name = self.class.to_s.split("::").last
    @uistyle = if UIStyle::Intertia.constants.include? class_name
      UIStyle::Intertia.const_get( class_name ) else UIStyle end
  end

  def contains_point( point )
    (point.x > 0 and point.x < width) and (point.y > 0 and point.y < height)
  end

  def view_at( origin )
    subviews.reverse_each do | view |
      child_origin = origin - view.origin
      if view.contains_point( child_origin )
        return view.view_at( child_origin )
      end
    end
    return self
  end  

  def screen_origin
    origin + if parent then parent.screen_origin else Point.ZERO end
  end

  def mouse_hover?()
    @mouse_hover
  end
  
  # ------------------------------------------------------------------------ #
  class Positioning; end
  # ------------------------------------------------------------------------ #
  
  def origin=( value )
    @origin = value.to_point
    origin_changed()
  end

  def extent=( value )
    temp = value.to_point()
    temp = Point.new( [temp.x, min_size.x].max(), [temp.y, min_size.y].max() )
    temp2 = Point.new( temp.x - width, temp.y - height )
    @extent = temp
    extent_changed( temp2.x, temp2.y )
  end

  include ViewRect

  def origin_changed()
  end

  def extent_changed( x, y )
    event = ShapeEvent.new( :view_size, Point.new( x, y ) )
    self.send_event( event )

    subviews.each do | child |
      #child.parent_resize( event.origin )
      delta = event.origin
      if child.reshape[0] == :move or child.reshape[1] == :move
        x = child.left + if child.reshape[0] == :move then delta.x else 0 end
        y = child.top + if child.reshape[1] == :move then delta.y else 0 end
        child.origin = [x, y]
      end
      if child.reshape[0] == :size or child.reshape[1] == :size
        w = child.width + if child.reshape[0] == :size then delta.x else 0 end
        h = child.height + if child.reshape[1] == :size then delta.y else 0 end
        child.extent = [w, h]
      end
    end
  end

  # ------------------------------------------------------------------------ #
  class Hierarchy; end
  # ------------------------------------------------------------------------ #

  def add_child( child )
    child.parent = self
    @subviews.push( child )
  end

  def remove_child( child )
    @subviews.delete( child )
    child.parent = nil
  end

  def bring_to_front()
    parent.subviews.delete( self )
    parent.subviews.push( self )
  end

  def send_to_back()
    parent.subviews.delete( self )
    parent.subviews[0, 0] = self
  end  
  
  # ------------------------------------------------------------------------ #
  class Drawing; end
  # ------------------------------------------------------------------------ #
  
  def draw()
    GL::PushMatrix()
      GL::Translate( @origin.x, @origin.y, 0 )
      GL::PushMatrix()
        uistyle.draw_content( self, @@graphics )
        draw_content()
      GL::PopMatrix()
      draw_children()
      uistyle.draw_border( self, @@graphics )
      draw_border()

      if $screen.meta_object == self
        GL::LineWidth( 2.0 )

        GL::Color( 0.0, 0.0, 0.0 ); 
        GL::Begin( GL::LINE_LOOP )
          GL::Vertex(         0 - 1.0,          0 - 1.0 )
          GL::Vertex(         0 - 1.0, 0 + height + 1.0 )
          GL::Vertex( 0 + width + 1.0, 0 + height + 1.0 )
          GL::Vertex( 0 + width + 1.0,          0 - 1.0 )
        GL::End()
      
        GL::Begin( GL::QUADS )
          GL::Vertex( width - 10, height - 10 )
          GL::Vertex( width - 10,      height )
          GL::Vertex( width,           height )
          GL::Vertex( width,      height - 10 )
        GL::End()
      end
    GL::PopMatrix()
  end
  
  # -------------------------------------------------------------------------- #
  
  def draw_content()
    GL::Color( 0.9375, 0.9375, 0.9375 ); 

    if @back_color
      GL::Color( back_color ); 
      GL::Begin( GL::QUADS )
        GL::Vertex(     0,      0 )
        GL::Vertex(     0, height )
        GL::Vertex( width, height )
        GL::Vertex( width,      0 )
      GL::End()
    end
  end

  # -------------------------------------------------------------------------- #

  def draw_children()
    @subviews.each do | child |
      child.draw()
    end
  end

  # -------------------------------------------------------------------------- #

  def draw_border()
    line_width = 0.5
    GL::LineWidth( 1 )
    GL::Color( 0.75, 0.75, 0.75, 1.0 ); 

    GL::Begin( GL::LINE_LOOP )
      GL::Vertex(     0 - line_width,      0 - line_width )
      GL::Vertex(     0 - line_width, height + line_width )
      GL::Vertex( width + line_width, height + line_width )
      GL::Vertex( width + line_width,      0 - line_width )
    GL::End()
  end

  # ------------------------------------------------------------------------ #
  class Events; end
  # ------------------------------------------------------------------------ #

  def send_event( event )
    if event.instance_of? MouseEvent
      @parent.focused_view = self if @parent and event.type == :mouse_down
      xevent = event.clone

      if !$screen.grabbed_view
        subviews.reverse_each do | child |
          if child.contains_point( event.origin - child.origin )
            return child.send_event( MouseEvent.new( event.type, event.button, event.origin - child.origin ) )
          end
        end
      end

      @mouse_hover = true if event.type == :mouse_enter
      @mouse_hover = false if event.type == :mouse_exit
      
      mouse_down( event.button, event.origin ) if event.type == :mouse_down
      mouse_up( event.button, event.origin ) if event.type == :mouse_up
      mouse_move( event.origin ) if event.type == :mouse_move
      mouse_drag( event.origin ) if event.type == :mouse_drag
      mouse_enter( event.origin ) if event.type == :mouse_enter
      mouse_exit( event.origin ) if event.type == :mouse_exit
      drag_accept( event.origin, event.object ) if event.type == :drag_accept
      drag_drop( event.object ) if event.type == :drag_drop
      #send( event.type, origin )

      return self
    end

    if event.instance_of? KeyEvent and event.type == :key_down
      #puts "#{self.class}.send_event( #{event} )"
      if @focused_view
        return @focused_view.send_event( event )
      end
      key_down( event.keysym, event.value ) if event.type == :key_down
    end

    view_size( event.origin ) if event.type == :view_size
    
    return self
  end

  alias :view_send_event :send_event
  
  include ViewEvents

  def draw_shadow()
    size, rnd, alpha = 7, 0.7, 0.125

    GL::PushMatrix()
      GL::Translate( 0.0, 1.0, 0.0 )
      GL::Begin( GL::QUADS )
        GL::Color( 0.0, 0.0, 0.0, alpha )
          GL::Vertex(     0,      0 )
          GL::Vertex(     0, height )
          GL::Vertex( width, height )
          GL::Vertex( width,      0 )
      GL::End()

      border_points = [
        [     0,      0 ], [     0 - size * rnd,      0 - size * rnd ],
        [     0,      0 ], [           0 - size,                   0 ],
        [     0, height ], [           0 - size,              height ],
        [     0, height ], [     0 - size * rnd, height + size * rnd ],
        [     0, height ], [                  0,       height + size ],
        [ width, height ], [          width + 0,       height + size ],
        [ width, height ], [ width + size * rnd, height + size * rnd ],
        [ width, height ], [       width + size,          height + 0 ],
        [ width,      0 ], [       width + size,                   0 ],
        [ width,      0 ], [ width + size * rnd,      0 - size * rnd ],
        [ width,      0 ], [              width,            0 - size ],
        [     0,      0 ], [                  0,            0 - size ],
        [     0,      0 ], [     0 - size * rnd,      0 - size * rnd ]
      ]

      GL::Begin( GL::QUAD_STRIP )
        border_points.each_slice( 2 ) do | point1, point2 |
          GL::Color( 0.0, 0.0, 0.0, alpha );  GL::Vertex( point1 )
          GL::Color( 0.0, 0.0, 0.0, 0.0 );    GL::Vertex( point2 )
        end
      GL::End()
    GL::PopMatrix()
  end

end # View

end
