#
# Copyright (c) 2006 Mike L. Austin.  All rights reserved.
# see LICENCE.txt for details
#

require 'view'
require 'font'
require 'events'

# ---------------------------------------------------------------------------- #

module Inertia

class Window < View

  @source = __FILE__
  @@shadow_next = 130

  attr :title, true
  attr :first_mouse
  attr :shadow_id
 
  alias :view_add_child :add_child
  
  # ------------------------------------------------------------------------ #
  class General; end
  # ------------------------------------------------------------------------ #

  def initialize()
    super
    window = self
    @@shadow_next += 1
    @shadow_id = @@shadow_next
    self.title = "<Untitled>"
    self.origin = [50, 50]
    self.extent = [300, 200]  # Race condition - may fire events before initialized
    self.min_size = Point.new( 150, 100 )

    view_add_child( @client = Client.new.with do | c |
      c.origin = [0, 25]; c.extent = [300, 200 - 25]
      c.reshape = [:size, :size]
    end )
    view_add_child(
      HoverButton.new.with do
        self.caption = "X"
        self.origin = [4, 4];  self.extent = [17, 17]
        self.bkg_color = [1.0, 0.75, 0.75]
        self.action = Action.new( window, :close_window, [] )
      end
    )
    view_add_child(
      HoverButton.new.with do
        self.caption = "T"
        self.origin = [300 - 4 - 17, 4];  self.extent = [17, 17]
        self.reshape = [:move, nil]
        self.bkg_color = [1.0, 0.75, 0.75]
        self.action = Action.new( window, :maximize_window, [] )
      end
    )
    view_add_child( Sizer.new.with do | s |
      s.origin = [300 - 25, 200 - 25]
    end )
  end

  def add_child( child )
    @client.add_child( child )
  end
  
  def close_window()
    parent.remove_child( self )
  end

  def maximize_window()
    parent.maximize_child( self )
  end

  # ------------------------------------------------------------------------ #
  class Events; end
  # ------------------------------------------------------------------------ #

  def send_event( event )
    bring_to_front() if event.type == :mouse_down
    return super
  end

  def mouse_down( button, origin )
    send_to_back() if button == 5
    @first_mouse = origin
  end

  def mouse_drag( origin )
    self.origin += origin - @first_mouse
  end

  def view_size( delta )
    if (@parent and @parent.focused_view != self) or (@parent == nil)
      GL::NewList( @shadow_id, GL::COMPILE )
        draw_shadow()
      GL::EndList()
    end
  end
  
  def focus_lost()
    GL::NewList( @shadow_id, GL::COMPILE )
      draw_shadow()
    GL::EndList()
  end

  # ------------------------------------------------------------------------ #
  class Drawing; end
  # ------------------------------------------------------------------------ #

  def draw_content()
    # Active windows may be risizing, so don't use the dlist
    if @parent and @parent.focused_view == self
      draw_shadow()
    else
      GL::CallList( @shadow_id )
    end

    # Draw background
    GL::Begin( GL::QUADS )
      GL::Color( 0.9375, 0.9375, 0.9375, 0.95 )
        GL::Vertex(     0, 100 )
        GL::Vertex( width, 100 )
      GL::Color( 1.0, 1.0, 1.0, 0.95 )
        GL::Vertex( width,      0 )
        GL::Vertex(     0,      0 )
      GL::Color( 0.9375, 0.9375, 0.9375, 0.95 )
        GL::Vertex(     0,    100 )
        GL::Vertex(     0, height )
        GL::Vertex( width, height )
        GL::Vertex( width,    100 )
    GL::End()

    # Draw title
    if @parent and @parent.focused_view == self
      # Title background
      GL::Begin( GL::QUADS )
        GL::Color( 0.8, 0.9, 1.0, 0.95 )
          GL::Vertex(         75, 12 )
          GL::Vertex( width - 75, 12 )
        GL::Color( 0.9, 0.95, 1.0, 0.95 )
          GL::Vertex( width - 75,  0 )
          GL::Vertex(         75,  0 )

        GL::Color( 0.8, 0.9, 1.0, 0.95 )
          GL::Vertex(         75, 25 )
          GL::Vertex( width - 75, 25 )
        GL::Color( 0.7, 0.85, 1.0, 0.95 )
          GL::Vertex( width - 75,  12 )
          GL::Vertex(         75,  12 )
      GL::End()

      # Title border
      line_width = 0.5
      GL::LineWidth( 1.0 )
      GL::Begin( GL::LINE_STRIP )
        GL::Color( 0.75, 0.75, 0.75, 1.0 )
          GL::Vertex(         75 - line_width,    - line_width )
          GL::Vertex(         75 - line_width, 24 + line_width )
          GL::Vertex(         76 - line_width, 25 + line_width )
          GL::Vertex( width - 76 + line_width, 25 + line_width )
          GL::Vertex( width - 75 + line_width, 24 + line_width )
          GL::Vertex( width - 75 + line_width,    - line_width )
      GL::End()
    end

    # Draw title
    GL::Color( 0.1, 0.1, 0.1, 1.0 )
    GL::Translate( width / 2 - $systemFont.text_length( title ) / 2,
                   25 / 2 - $systemFont.font.height / 2, 0.0 )
    $systemFont.draw_string( title )
  end

  # ------------------------------------------------------------------------ #

  def draw_border()
    half_width = 1
    GL::LineWidth( 2.0 )
    GL::Enable( GL::LINE_SMOOTH )
    #GL::Color( 0.75, 0.75, 0.75, 1.0 )
    GL::Color( 0.75, 0.75, 0.75, 1.0 )
    GL::Begin( GL::LINE_LOOP )
      GL::Vertex(     0 - half_width,      0 - half_width )
      GL::Vertex(     0 - half_width, height + half_width )
      GL::Vertex( width + half_width, height + half_width )
      GL::Vertex( width + half_width,      0 - half_width )
    GL::End()
  end

  # ------------------------------------------------------------------------ #


end # Window

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #

class Client < View

  def mouse_down( button, origin )
    @first_mouse = origin
  end

  def mouse_drag( origin )
    parent.origin += origin - @first_mouse
  end
  
  def draw_content()
  end
  
  def draw_border()
  end
  
  def draw_children()
    sorigin = screen_origin
    GL::PushAttrib( GL::SCISSOR_BIT )
    GL::Scissor( sorigin.x, $screen.height - sorigin.y - extent.y, extent.x, extent.y )
    GL::Enable( GL::SCISSOR_TEST )
    super
    GL::Disable( GL::SCISSOR_TEST )
    GL::PopAttrib()
  end
    
end # Client

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #

class Sizer < View

  attr :first_mouse

  # -------------------------------------------------------------------------- #
  class General; end
  # -------------------------------------------------------------------------- #

  def initialize( title = "<Untitled>" )
    super()
    self.extent = [25, 25]
    self.reshape = [:move, :move]
  end

  # -------------------------------------------------------------------------- #
  class Events; end
  # -------------------------------------------------------------------------- #

  def mouse_down( button, origin )
    @first_mouse = origin
  end

  def mouse_drag( origin )
    parent.extent += origin - @first_mouse
  end

  # -------------------------------------------------------------------------- #
  class Drawing; end
  # -------------------------------------------------------------------------- #

  def draw_content()
    GL::Color( 0.75, 0.75, 0.75 )
    GL::LineWidth( 1.0 )
    GL::Begin( GL::LINES )
      GL::Vertex( 3,  height - 3 );  GL::Vertex( width - 3,  3 )
      GL::Vertex( 8,  height - 3 );  GL::Vertex( width - 3,  8 )
      GL::Vertex( 13, height - 3 );  GL::Vertex( width - 3, 13 )
      GL::Vertex( 18, height - 3 );  GL::Vertex( width - 3, 18 )
    GL::End()
  end

  def draw_border()
  end

end # Sizer

end
